import pickle
import datetime
import networkx as nx
import pandas as pd
import numpy as np

from .utils import (
    analyze_neighborhood_attributes,
    print_neighbors_prob,
    heat_map_prob,
    plot_distribution,
    plot_community_composition,
    matplotlib_graph_visualization
)

def analyze_graph(graph_path=None, 
                  attribute=None, 
                  clustering_method='hierarchical', 
                  inconsistency_threshold=0.1, 
                  verbose=True,
                  plot_graph=False, 
                  neigh_prob_path = None,
                  degree_distribution_outpath = None,
                  betweenness_distribution_outpath = None,
                  prob_heatmap_path = None,
                  community_composition_outpath = None,
                  graph_visualization_path = None):
    if isinstance(graph_path, str):
        G = pickle.load(open(graph_path, 'rb'))
    elif isinstance(graph_path, nx.Graph):
        G = graph_path
    else:
        raise ValueError("Invalid graph_path. Must be a path to a file or a NetworkX Graph.")

    if neigh_prob_path is None:
        neigh_prob_path = f"./neighbor_stat_{datetime.datetime.now().strftime('%Y%m%d%H%M')}.dat"

    if verbose:
        print(f"--------------------------\nGraph analysis options\n--------------------------\n\n"
              f"\tOptions:\n"
              f"\tgraph_path: {graph_path}, attribute: {attribute}, \n"
              f"\tclustering_method: {clustering_method}, inconsistency_threshold: {inconsistency_threshold}, verbose: {verbose}\n\n")

    if attribute is not None:
        df_neigh = analyze_neighborhood_attributes(G, attribute_name = attribute)
        probabilities = print_neighbors_prob(df_neigh, attribute)
        for (i, j), prob in probabilities.items():
            print(f"P({j}|{i}) = {prob}")
        with open(neigh_prob_path, 'w') as fp:
            for (i, j), prob in probabilities.items():
                fp.write(f"P({j}|{i}) = {prob}")   

        heat_map_prob(probabilities, df_neigh, attribute, prob_heatmap_path)

    degree_data = {'data': [degree for _, degree in G.degree()],
                   'title': 'Degree distribution',
                   'xlabel': 'Degree',
                   'ylabel': 'Number of Nodes'}
    betweenness_data = {'data': [degree for _, degree in G.degree()],
                   'title': 'Degree distribution',
                   'xlabel': 'Degree',
                   'ylabel': 'Number of Nodes'}
    
    plot_distribution(degree_data, degree_distribution_outpath)
    plot_distribution(betweenness_data, betweenness_distribution_outpath)
    plot_community_composition(G, attribute, community_composition_outpath)
    if plot_graph:
        matplotlib_graph_visualization(G, attribute, graph_visualization_path, palette = 'viridis')



import datetime
import os
import numpy as np
import pandas as pd
import pickle
import networkx as nx
from scipy.spatial.distance import pdist, squareform
from sklearn.metrics.pairwise import cosine_similarity

def create_graph(input_dataframe=None, preprocessed_dataframe=None, output_path=None, method='knn', threshold=0.75, k=5, verbose=True):
    if verbose:
        print(f"--------------------------\nGraph creation options\n--------------------------\n\n"
              f"\tOptions:\n"
              f"\input_dataframe: {input_dataframe}, preprocessed_dataframe: {preprocessed_dataframe}, \n"
              f"\toutput_path: {output_path}, method: {method}, threshold: {threshold}, k: {k}, verbose: {verbose}\n\n")

    # Load dataframe
    if isinstance(input_dataframe, str):
        df = pd.read_pickle(input_dataframe) if input_dataframe.endswith('.pickle') else pd.read_csv(input_dataframe)
    elif isinstance(input_dataframe, pd.DataFrame):
        df = input_dataframe.copy()
    else:
        raise ValueError("Invalid input_dataframe. Must be a path to a file or a pandas DataFrame.")

    if preprocessed_dataframe is None:
        df_preprocessed = df.copy()
    else:
        if isinstance(preprocessed_dataframe, str):
            df_preprocessed = pd.read_pickle(preprocessed_dataframe) if preprocessed_dataframe.endswith('.pickle') else pd.read_csv(preprocessed_dataframe)
        elif isinstance(preprocessed_dataframe, pd.DataFrame):
            df_preprocessed = preprocessed_dataframe.copy()
        else:
            raise ValueError("Invalid preprocessed_dataframe. Must be a path to a file or a pandas DataFrame.")

    if df.shape[0] != df_preprocessed.shape[0]:
        df_preprocessed = df.dropna().copy()
        if verbose:
            print(f"{datetime.datetime.now()}: Dropped rows with NaN values from the original dataframe due to mismatch with preprocessed dataframe.")

    G = nx.Graph()

    for i, row in df.iterrows():
        G.add_node(i, **row.to_dict())

    if method == 'knn':
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        values = df_preprocessed.select_dtypes(include=numerics).values
        dists = squareform(pdist(values, metric='euclidean'))
        for i in range(len(df)):
            knn_indices = np.argsort(dists[i])[:k + 1]
            for j in knn_indices:
                if i != j:
                    G.add_edge(i, j)
    elif method == 'distance_threshold':
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        values = df_preprocessed.select_dtypes(include=numerics).values
        dists = squareform(pdist(values, metric='euclidean'))
        for i in range(len(df)):
            for j in range(i + 1, len(df)):
                if dists[i, j] <= threshold:
                    G.add_edge(i, j)
    elif method == 'similarity':
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        values = df_preprocessed.select_dtypes(include=numerics).values
        sim_matrix = cosine_similarity(values)
        for i in range(len(df)):
            for j in range(i + 1, len(df)):
                if sim_matrix[i, j] >= threshold:
                    G.add_edge(i, j)
    else:
        raise ValueError(f"Unsupported method: {method}")

    if output_path is None:
        if isinstance(input_dataframe, str):
            base, _ = os.path.splitext(input_dataframe)
            output_path = f"{base}_graph_{datetime.datetime.now().strftime('%Y%m%d%H%M')}.pickle"
        else:
            output_path = f"./graph_{datetime.datetime.now().strftime('%Y%m%d%H%M')}.pickle"

    pickle.dump(G, open(output_path, 'wb'))
    if verbose:
        print(f"{datetime.datetime.now()}: Nodes attributes:\n" 
              f"{G.nodes[0].keys()}")
        print(f"{datetime.datetime.now()}: Saved graph to {output_path}.")

    return G



import json
import os

default_config = {
    "nan_action": "infer",
    "nan_threshold": 0.6,
    "numeric_scaling": "standard",
    "categorical_encoding": "one-hot",
    "manifold_method": None,
    "manifold_dim": 2,
    "verbose": True,
    "method": "knn",
    "threshold": 0.75,
    "k": 5,
    "clustering_method": "hierarchical",
    "inconsistency_threshold": 0.1
}

def load_config(config_path="config.json"):
    
    if type(config_path) is str:
        if os.path.exists(config_path):
            with open(config_path, 'r') as f:
                config = json.load(f)
            if config["verbose"]:
                print(f"Loaded configuration from {config_path}.")
            return config
    else:
        config = default_config
        if config["verbose"]:
            print("Using default configuration.")
    return config

def save_config(config, config_path="config.json"):
    with open(config_path, 'w') as f:
        json.dump(config, f, indent=4)
    if config["verbose"]:
        print(f"Configuration saved to {config_path}.")
